
import java.util.*;

public final class Package implements Comparable {

   public static Hashtable heldPackages = new Hashtable();
   public static Hashtable freePackages = new Hashtable();
   public static Hashtable pkgsByLocation = new Hashtable();
   public static Hashtable maybePackages = new Hashtable();
   
   public String id;
   public Point loc;
   public Point dest;  //null -> maybePackage
   public int weight;
   public boolean held = false;
   public boolean tracking = false;
   
   public Package(String id) {
      this.id = id;
   }

   //mark this package for retrieval by location since it is available
   //for pickup
   public final void track() {
      if (tracking) return;  //already tracking
      if (loc == null) return; //can't track if we don't know the loc
      TreeSet h1 = (TreeSet) pkgsByLocation.get(loc);
      if (h1 == null) {
         h1 = new TreeSet();
         pkgsByLocation.put(loc, h1);
      }
      h1.add(this);
      tracking = true;
      freePackages.put(id, this);
   }
   
   //do not follow closely the locations of this package as it is
   //being held by a robot
   public final void untrack() {
      if (!tracking) return; //can't be tracking this one
      if (loc == null) return; //can't be tracking this one
      TreeSet h1 = (TreeSet) pkgsByLocation.get(loc);
      if (h1 != null) {
         h1.remove(this);
         if (h1.isEmpty()) pkgsByLocation.remove(loc);
      }
      tracking = false;
      freePackages.remove(id);
      heldPackages.put(id, this);
   }
   
   public static final void reportedPackage(String id, int dx, int dy, int w, Point l) {
      Package p = null;
      if (!freePackages.containsKey(id)) { //if this is a new package
         if ((p = (Package)maybePackages.get(id)) != null) {
            maybePackages.remove(id);
         }
         else {
            p = new Package(id);
         }
         p.weight = w;
         p.dest = new Point(dx, dy);
         p.loc = new Point(l.x, l.y);
         p.track();  //this will add it to freePackages
      }
   }
   
   public static final Package maybePackage(String id, Point l) {
      Package p = (Package) maybePackages.get(id);
      if (p == null) {
         p = new Package(id);
         maybePackages.put(id, p);
      }
      p.loc = l;
      return p;
   }

   public final boolean equals(Object pkg) {
      try {
         if (id.equals(((Package)pkg).id)) return true;
      } catch (Exception ex) {
      }
      return false;
   }

   public final int compareTo(Object pkg) {
      if (equals(pkg)) return 0;
      Package p = (Package)pkg;
      if (weight == p.weight) {
         int d1 = loc.distance(dest);
         int d2 = p.loc.distance(p.dest);
         if (d1 == d2) {   
            return -id.compareTo(p.id);
         }
         else {
            return d1 - d2; //ascending on distance to dest
         }
      }
      return p.weight - weight; //this puts weights in descending order
   }
   
   public static final Iterator getPackageList(Point loc) {
      TreeSet t = (TreeSet) pkgsByLocation.get(loc);
      return t == null ? null : t.iterator();
   }
   
   public final void disappear() {
      //only packages that were held can possibly disappear
      //we are not tracking any package that was held
      heldPackages.remove(id);
      freePackages.remove(id); //should never be here, but just in case
      maybePackages.remove(id);
   }
   
   //closest package being held, including maybe packages
   public static Point closestHeldPackage(Robot r) {
      int minDist = 2001;  //longest distance between pts on largest board
      Point result = null;
      for (Iterator i = Robot.allRobots.values().iterator(); i.hasNext(); ) {
         Robot b = (Robot) i.next();
         if (b == r) continue;
         if (b.myPackages.isEmpty()) continue;
         int dist = r.loc.distance(b.loc);
         if (dist < minDist) {
            result = b.loc;
            minDist = dist;
         }
      }
      return result;
   }
   
   //this will be a dropped maybe package because closestHeldPackage
   //handles held maybe packages
   public static Point closestPseudoPackage(Point p) {
      int minFreeDist = 2001;  //longest distance between pts on largest board
      Point freeResult = null;
      for (Iterator i = maybePackages.values().iterator(); i.hasNext(); ) {
         Package pkg = (Package)i.next();
         if (!pkg.held) {
            Point h = pkg.loc;
            int dist = p.distance(h);
            if (dist < minFreeDist) {
               freeResult = h;
               minFreeDist = dist;
            }
         }
      }
      return freeResult;
   }
}